
package RightFieldsApp;
use strict;

use base qw(MT::App::CMS);

use RightFields qw( load_plugindata save_plugindata config_key
	mt_fields mt_field_order load_rf_settings save_rf_settings current_ts );

use Data::Dumper;
my @standard_elems = qw( label type rows width length choices weblog 
	upload_path url_path overwrite date_order date_y_start date_y_end
	date_show_hms date_time	date_minutes date_ampm
	filenames category_ids show_buttons text_filters );

my $new_rows = 3;

sub init {
	my $app = shift;
	$app->SUPER::init (@_) or return;
	$app->add_methods (
		'list_blogs' => \&list_blogs,
		'cfg_standard' => \&cfg_standard,
		'cfg_extra' => \&cfg_extra,
		'cfg_defaults' => \&cfg_defaults,
		'cfg_tags' => \&cfg_tags,
		'cfg_listing' => \&cfg_listing,
		'save_cfg' => \&save_cfg,
		'delete' => \&delete_confirm,
		'do_delete' => \&do_delete,
		'override' => \&override
	);
	$app->{'default_mode'} = 'list_blogs';
	$app->{'user_class'} = 'MT::Author';
	$app->{'plugin_template_path'} = 'tmpl';
	$app->{'requires_login'} = 1;
	$app;
}

sub can_edit_blog {
	my ($author, $blog_id) = @_;
	return 1 if $author->is_superuser();
	my $perms = MT::Permission->load({ 'author_id' => $author->id,
		'blog_id' => $blog_id });
	return ($perms && $perms->can_administer_blog) ? 1 : 0;
}

sub permission_check {
	my ($author, $blog_id) = @_;
	if ($blog_id == -1) {
		if (!$author->is_superuser) {
			die 'You must be a System Administrator to change Default Settings.';
		}
	} else {
		if (!can_edit_blog($author, $blog_id)) {
			die 'You do not have permission to change settings for that weblog.';
		}
	}
}

sub build_page {
	my ($app, $page, $param) = @_;
	my $rightfields = MT::Plugin::RightFields->instance;
	my $q = $app->{'query'};
	$param->{'blog_loop'} = blogs_for_author($app, $q->param('blog_id'));
	if (!@{$param->{'blog_loop'}}) {
		$param->{'no_perms'} = 1;
	}
	if ($q->param('blog_id') && $app->{'author'}) {
		if (!can_edit_blog($app->{'author'}, $q->param('blog_id'))) {
			die 'You do not have Administrator permissions for that weblog.';
		}
		$param->{'rf_blog_id'} = $param->{'default'} ? -1
			: $q->param('blog_id');
	}
	if ($app->{'author'}) {
		$param->{'is_superuser'} = $app->{'author'}->is_superuser();
	}
	for my $key (qw( saved deleted overridden )) {
		$param->{$key} = $q->param($key);
	}
	$param->{'system_overview_nav'} = (!$q->param('blog_id') || ($q->param('blog_id') < 1));
	$param->{'default'} = ($q->param('blog_id') && ($q->param('blog_id') == -1));
	$app->SUPER::build_page($page, $param);
}

sub list_blogs {
	my ($app) = @_;
	my $q = $app->{'query'};
	my $param = {};
	$app->add_breadcrumb('RightFields');
	$param->{'ef_converted'} = $q->param('ef_converted');
	$param->{'can_convert'} = $app->{'author'}->is_superuser();
	$param->{'menu_mode'} = 'cfg_standard';
	return $app->build_page('rf_menu.tmpl', $param);
}

sub cfg_standard {
	my ($app) = @_;
	my $q = $app->{'query'};
	my $param = {};
	my $data = load_rf_settings($q, 'standard') || {};
	$param->{'using_default'} = using_default($q, 'standard', $data);
	if (!$param->{'using_default'}) {
		my $fields = mt_fields();
		my $order = mt_field_order();
		$data->{'enabled'} ||= 1;
		if ($data->{'enabled'} == -1) {
			$param->{'cust_disabled'} = 1;
		} elsif ($data->{'enabled'} == 2) {
			$param->{'cust_after'} = 1;
		} else {
			$param->{'cust_enabled'} = 1;
		}
		$param->{'enabled_ts'} = $data->{'enabled_ts'} ||
			current_ts($q->param('blog_id'));
		$param->{'enable_label'} = 'Customization is:';
		$param->{'field_loop'} = std_field_loop($data, $fields, $order);
	}
	$param->{'tab_standard'} = ' class="yah"';
	$param->{'menu_mode'} = 'cfg_standard';
	$app->add_breadcrumb('RightFields Standard Field Configuration');
	return $app->build_page('rf_cfg_standard.tmpl', $param);
}

sub ext_field_loop {
	my ($q) = @_;
	my $ext_data = load_rf_settings($q, 'extra');
	if (!$ext_data) {
		$ext_data = load_plugindata(config_key(-1, 'extra')) || {};
	}
	my $ext_field_loop = [];
	if ($ext_data->{'datasource'}) {
		if ($ext_data->{'datasource'} eq '_pseudo') {
			$ext_field_loop = pseudo_field_loop($ext_data, 1);
		} else {
			my $db_defs = db_defs($ext_data->{'datasource'});
			$ext_field_loop = sql_field_loop($ext_data, $db_defs);
		}
	}
	return $ext_field_loop;
}

sub cfg_extra {
	my ($app) = @_;
	my $q = $app->{'query'};
	my $param = {};
	my $data = load_rf_settings($q, 'extra') || {};
	$param->{'using_default'} = using_default($q, 'extra', $data);
	if (!$param->{'using_default'}) {	
		if ($data->{'datasource'}) {
			if ($data->{'datasource'} eq '_pseudo') {
				$param->{'pseudo'} = 1;
				$param->{'field_loop'} = pseudo_field_loop($data);
				$param->{'field_keys'} = join(',', keys %{$data->{'cols'}});
				$param->{'new_rows'} = $new_rows;
			} else {
				my $db_defs = db_defs($data->{'datasource'});
				if (!ref($db_defs)) {
					$param->{'bad_datasource'} = $db_defs;
				} else {
					$param->{'field_loop'} = sql_field_loop($data, $db_defs);
					$param->{'field_keys'} = join(',', keys %$db_defs);
				}
			}
		} else {
				# default
			$param->{'pseudo'} = 1;
		}
		$data->{'enabled'} ||= 1;
		if ($data->{'enabled'} == -1) {
			$param->{'cust_disabled'} = 1;
		} elsif ($data->{'enabled'} == 2) {
			$param->{'cust_after'} = 1;
		} else {
			$param->{'cust_enabled'} = 1;
		}
		$param->{'enabled_ts'} = $data->{'enabled_ts'} ||
			current_ts($q->param('blog_id'));
		$param->{'enable_label'} = 'Extra fields are:';
		$param->{'datasource'} = $data->{'datasource'};
	}
	$param->{'tab_extra'} = ' class="yah"';
	$param->{'menu_mode'} = 'cfg_extra';
	$param->{'default'} = ($q->param('blog_id') == -1);
	$app->add_breadcrumb('RightFields Extra Field Configuration');
	return $app->build_page('rf_cfg_extra.tmpl', $param);
}

sub cfg_defaults {
	my ($app) = @_;
	my $q = $app->{'query'};
	my $param = {};
	my $defaults = load_rf_settings($q, 'defaults') || {};
	$param->{'using_default'} = using_default($q, 'defaults', $defaults);
	if (!$param->{'using_default'}) {
		my $fields = mt_fields();
		my $order = mt_field_order();
		my @field_keys = ();
		my $std_data = load_rf_settings($q, 'standard') || {};
		my $std_field_loop = std_field_loop($std_data, $fields, $order);
		for my $field (@$std_field_loop) {
			$field->{'standard'} = 1;
		}
		my $ext_field_loop = ext_field_loop($q);
		$param->{'field_loop'} = [];
		push(@{$param->{'field_loop'}}, @$std_field_loop, @$ext_field_loop);
		for my $field (@{$param->{'field_loop'}}) {
			$field->{"type_$field->{'type'}"} = 1;
			$field->{'display_key'} = $field->{'field_key'};
			if ($field->{'standard'}) {
				$field->{'field_key'} = 'std_' . $field->{'field_key'};
			}
			$field->{'value'} = $defaults->{$field->{'field_key'}} || '';
			if ($field->{'type'} =~ /^text/) {
				my $key = "$field->{'field_key'}_is_guide";
				$field->{'is_guide'} = $defaults->{$key};
				push(@field_keys, $key);
			} elsif ($field->{'type'} =~ /^menu|radio$/) {
				my @choices_loop;
				for my $choice (split(/[\n\r]+/, $field->{'choices'})) {
					next unless $choice;
					my ($value, $label) = split(/=/, $choice);
					$label ||= $value;
					push(@choices_loop, {
						'value' => $value,
						'selected' => ($field->{'value'} && 
							(($field->{'value'} eq $value) || 
								$field->{'value'} eq $label)) ? 1 : 0,
						'label' => $label
					});
				}
				$field->{'choices_loop'} = \@choices_loop;
			} elsif ($field->{'type'} eq 'date') {
				if ($field->{'value'} eq 'current') {
					$field->{'date_rb_current'} = 1;
					$field->{'value'} = '';
				} elsif ($field->{'value'}) {
					$field->{'date_rb_ts'} = 1;
				} else {
					$field->{'date_rb_none'} = 1;
				}
			} elsif ($field->{'type'} eq 'entry') {
				next unless ($field->{'weblog'});
				my @entry_rows = ();
				my @entries = MT::Entry->load({
					'blog_id' => $field->{'weblog'}
				});
				for my $entry (sort { $a->title cmp $b->title } @entries) {
					push (@entry_rows, {
						'id' => $entry->id,
						'title' => (length($entry->title) > 80)
							? substr($entry->title, 0, 80) . '...'
							: $entry->title,
						'selected' => (($field->{'value'} || 0) == $entry->id)
					});
				}
				$field->{'entries'} = \@entry_rows;
			}
			push(@field_keys, $field->{'field_key'});
		}
		$param->{'field_keys'} = join(',', @field_keys);
	}
	$param->{'tab_defaults'} = ' class="yah"';
	$param->{'menu_mode'} = 'cfg_defaults';
	$param->{'default'} = ($q->param('blog_id') == -1);
	$app->add_breadcrumb('RightFields Default Values Configuration');
	return $app->build_page('rf_cfg_defaults.tmpl', $param);
}

sub cfg_tags {
	my ($app) = @_;
	my $q = $app->{'query'};
	my $param = {};
	my $data = load_rf_settings($q, 'tags');
	$param->{'using_default'} = using_default($q, 'tags', $data);
	if (!$param->{'using_default'}) {
		$data ||= [];
		for (1 .. $new_rows) {
			push(@$data, { 'tag' => '', 'field' => '' });
		}
			# get standard field list
		my $std_data = load_rf_settings($q, 'standard');
		if (!$std_data) {
			$std_data = load_plugindata(config_key(-1, 'standard')) || {};
		}
		my $fields = mt_fields();
		my $order = mt_field_order();
		my $std_field_loop = std_field_loop($std_data, $fields, $order);
			# prefix with 'std ' in case of an extra field with same column name
		for my $field (@$std_field_loop) {
			$field->{'display_key'} = $field->{'field_key'};
			$field->{'field_key'} = 'std ' . $field->{'field_key'};
		}
		my $ext_field_loop = ext_field_loop($q);
			# get extra field list
		my $i = 0;
		for my $row (@$data) {
			$row->{'field_loop'} = [];
			for my $field (@$std_field_loop, @$ext_field_loop) {
				push(@{$row->{'field_loop'}}, {
					'field_key' => $field->{'field_key'},
					'display_key' => $field->{'display_key'}
						|| $field->{'field_key'},
					'label' => $field->{'label'},
					'selected' => ($field->{'field_key'} eq $row->{'field'})
				});
			}
			$row->{'count'} = ++$i;
		}
		$param->{'tag_loop'} = $data;
		$param->{'tag_count'} = $i;
	}
	$param->{'tab_tags'} = ' class="yah"';
	$param->{'menu_mode'} = 'cfg_tags';
	$app->add_breadcrumb('RightFields Template Tags Configuration');
	return $app->build_page('rf_cfg_tags.tmpl', $param);
}

sub db_defs {
	my ($datasource) = @_;
	my %mt_tables = map { $_ => 1 } qw( author blog category comment entry fileinfo ipbanlist log notification permission placement plugindata session tbping template templatemap trackback config );
	if ($mt_tables{$datasource}) {
		return qq{Datasource cannot be one of Movable Type's own tables.};
	} else {
		my $holder = new DatasourceHolder;
		DatasourceHolder::datasource($datasource);
		my $driver = MT::Object->driver;
		my $db_defs = $driver->column_defs($holder);
		if (!$db_defs) {
			return 'Could not load datasource. Check table name.';
		} elsif (!$db_defs->{'id'}) {
			return qq{Table must have a "${datasource}_id" column.};
		} else {
			delete($db_defs->{'id'});
			return $db_defs;
		}
	}
}

sub cfg_listing {
	my ($app) = @_;
	my $q = $app->{'query'};
	my $param = {};
	my $data_key = config_key($q->param('blog_id'), 'listing');
	my $data = load_plugindata($data_key) || {};
	$param->{'using_default'} = using_default($q, 'listing', $data);
	if (!$param->{'using_default'}) {
		for my $key (qw( customize_compact compact expanded customize_expanded )) {
			$param->{$key} = $data->{$key};
		}
	}
	$param->{'tab_listing'} = ' class="yah"';
	$param->{'menu_mode'} = 'cfg_listing';
	$param->{'default'} = ($q->param('blog_id') == -1);
	$app->add_breadcrumb('RightFields Entry Listing Configuration');
	return $app->build_page('rf_cfg_listing.tmpl', $param);
}

sub save_cfg {
	my ($app) = @_;
	my $q = $app->{'query'};
	my $data = {};
	my $save = 1;
	my $type = $q->param('type');
	permission_check($app->{'author'}, $q->param('blog_id'));
	if ($type eq 'standard') {
		my $fields = mt_fields();
		for my $key (keys %$fields) {
			$data->{$key} = {};
			for my $elem ('customize', @standard_elems) {
				$data->{$key}->{$elem} = $q->param("rf_${elem}_${key}");
			}
			$data->{$key}->{'show_buttons'} ||= 0;
			$data->{$key}->{'text_filters'} ||= 0;
		}
		$data->{'enabled'} = $q->param('enabled');
		if ($data->{'enabled'} == 2) {
			$data->{'enabled_ts'} = $q->param('enabled_ts');
		}
	} elsif ($type eq 'datasource') {
		my $datasource = ($q->param('dstype') eq 'pseudo')
			? '_pseudo' : $q->param('datasource');
		my $old_data = load_rf_settings($q, 'extra') || {};
			# if for some reason they've resaved it with the same datasource,
			# leave the existing config alone
		if ($old_data->{'datasource'} eq $datasource) {
			$save = 0;
		} else {
			$data->{'datasource'} = $datasource;
		}
			# for redirect
		$type = 'extra';
	} elsif ($type eq 'extra') {
		my $old_data = load_rf_settings($q, 'extra') || {};
		$data->{'datasource'} = $old_data->{'datasource'};
		if (!$data->{'datasource'}) {
			$save = 0;
		} else {
			$data->{'enabled'} = $q->param('enabled');
			if ($data->{'enabled'} == 2) {
				$data->{'enabled_ts'} = $q->param('enabled_ts');
			}
			$data->{'cols'} = {};
			for my $key (split(/,/, $q->param('field_keys'))) {
				my $new_fieldname = $q->param("rf_fieldname_$key");
				if (defined($new_fieldname)) {
					next unless $new_fieldname;
					$key = $new_fieldname;
				}
				$data->{'cols'}->{$key} = {}; 
				for my $elem ('sort', 'ignore', @standard_elems) {
					$data->{'cols'}->{$key}->{$elem} = 
						$q->param("rf_${elem}_${key}");
				}
				$data->{'cols'}->{$key}->{'show_buttons'} ||= 0;
				$data->{'cols'}->{$key}->{'text_filters'} ||= 0;
			}
			for my $i (1 .. $q->param('new_rows')) {
				my $key = $q->param("rf_fieldname_new_$i");
				next unless $key;
				$data->{'cols'}->{$key} = {};
				for my $elem ('sort', @standard_elems) {
					$data->{'cols'}->{$key}->{$elem} = 
						$q->param("rf_${elem}_new_$i");
				}
				$data->{'cols'}->{$key}->{'show_buttons'} ||= 0;
				$data->{'cols'}->{$key}->{'text_filters'} ||= 0;
			}
		}
	} elsif ($type eq 'defaults') {
		for my $key (split(/,/, $q->param('field_keys'))) {
				# date field
			if ($q->param($key . '_rb') && ($q->param($key . '_rb') eq 'current')) {
				$q->param($key, 'current');
			}
			next unless ($q->param($key));
			$data->{$key} = $q->param($key);
		}
	} elsif ($type eq 'tags') {
		$data = [];
		for my $i (1 .. $q->param('tag_count')) {
			next unless $q->param("tag_$i");
			push(@$data, { 'tag' => $q->param("tag_$i"),
				'field' => $q->param("field_$i") });
		}	
	} elsif ($type eq 'listing') {
		$data = {};
		for my $key (qw( customize_compact compact expanded customize_expanded )) {
			$data->{$key} = $q->param($key);
		}
	}
	if ($save) {
		save_rf_settings($q, $type, $data)
			|| die "Couldn't save plugindata";
	}
	$app->redirect($app->uri( 'mode' => "cfg_$type",
		args => { 'blog_id' => $q->param('blog_id'), 'saved' => 1 } ));
}

sub override {
# override default settings: we just need to create an empty PluginData
# record for this blog ID and type
	my ($app) = @_;
	my $q = $app->{'query'};
	my $type = $q->param('type');
	my $data = ($type eq 'tags') ? [ ] : { 'dummy' => 1 };
	permission_check($app->{'author'}, $q->param('blog_id'));
	my $data_key = config_key($q->param('blog_id'), $type);
	save_plugindata($data_key, $data) || die "Couldn't save plugindata";
	$app->redirect($app->uri( 'mode' => 'cfg_' . $type,
		args => { 'blog_id' => $q->param('blog_id'), 'overridden' => 1 } ));
}

sub std_field_loop {
	my ($data, $fields, $order, $want_values) = @_;
	my @fields = ();
	for my $key (sort { $order->{$a} <=> $order->{$b} } keys %$fields) {
		my $row = { 'field_key' => $key, 'standard_label' => $fields->{$key} };
		if ($data && %$data && $data->{$key}) {
			for my $elem ('customize', @standard_elems) {
				$row->{$elem} = $data->{$key}->{$elem};
			}
		} else {
				# defaults
			$row->{'label'} = $fields->{$key};
			$row->{'type'} = ($key eq 'title') ? 'text' : 'textarea';
			$row->{'width'} = ($key eq 'title') ? 300 : 577;
			if ($key eq 'text_more') {
				$row->{'rows'} = 15;
			} elsif ($key eq 'excerpt') {
				$row->{'rows'} = 5;
			} elsif ($key eq 'keywords') {
				$row->{'rows'} = 2;
			} else {
				$row->{'rows'} = 10;
			}
		}
		row_defaults($row, $key);
		row_params($row);
		push(@fields, $row);
	}
	return \@fields;
}

sub delete_confirm {
	my ($app) = @_;
	my $param = {};
	my $blog_id = $app->{'query'}->param('blog_id');
	if ($blog_id == -1) {
		$param->{'default'} = 1;
	} else {
		$param->{'blog_id'} = $blog_id;
	}
	for my $type (qw( standard extra tags )) {
		my $data_key = config_key($blog_id, $type);
		if (load_plugindata($data_key)) {
			$param->{"has_$type"} = 1;
			$param->{'has_settings'} = 1;
		}
	}
	return $app->build_page('rf_delete_confirm.tmpl', $param);
}

sub do_delete {
	my ($app) = @_;
	my $q = $app->{'query'};
	if ($q->param('cancel')) {
		return list_blogs(@_);
	}
	my $blog_id = $q->param('blog_id');
	permission_check($app->{'author'}, $blog_id);
	require MT::PluginData;
	for my $type (qw( standard extra tags )) {
		if ($q->param("delete_$type")) {
			my $data_key = config_key($blog_id, $type);
			my $data = MT::PluginData->load({
				'plugin' => 'rightfields', 'key' => $data_key
			});
			$data->remove;
		}
	}
	$app->redirect($app->uri( 'mode' => "cfg_standard",
		args => { 'blog_id' => $blog_id, 'deleted' => 1 } ));
}

sub pseudo_field_loop {
	my ($data, $no_blanks) = @_;
	my @fields = ();
	my $count = scalar keys %{$data->{'cols'}};
	my @keys = sort {
		($data->{'cols'}->{$a}->{'sort'} || $count) <=>
		($data->{'cols'}->{$b}->{'sort'} || $count)
	} keys %{$data->{'cols'}};
	my $i = 0;
	for my $key (@keys, $no_blanks ? () : (1 .. $new_rows)) {
		$i++;
		my $row;
		if ($i > scalar @keys) {
			$row = { 'field_key' => "new_$key", 'standard_label' => '',
				'sort' => $i, 'label' => '', 'new' => 1, 'customize' => 0
			};
		} else {
			$row = { 'field_key' => $key, 'standard_label' => $key,
				'sort' => $i
			};
			for my $elem (@standard_elems) {
				$row->{$elem} = $data->{'cols'}->{$key}->{$elem};
			}
			$row->{'customize'} = 1;
		}
		row_defaults($row, $key);
		row_params($row);
		push(@fields, $row);
	}
	return \@fields;
}

sub sql_field_loop {
	my ($data, $db_defs) = @_;
	my @fields = ();
	my @keys;
	if ($data && %$data && $data->{'cols'}) {
		@keys = keys %$db_defs;
		
		@keys = sort {
			($data->{'cols'}->{$a}->{'sort'} || scalar @keys) <=>
			($data->{'cols'}->{$b}->{'sort'} || scalar @keys)
		} @keys;
	} else {
		@keys = sort keys %$db_defs;
	}
	my $i = 0;
	for my $key (@keys) {
		next if ($key eq 'id');
		my $row = { 'field_key' => $key, 'standard_label' => $key };
		$row->{'sort'} = ++$i;
		if ($data && %$data && $data->{'cols'}) {
			for my $elem ('ignore', @standard_elems) {
				$row->{$elem} = $data->{'cols'}->{$key}->{$elem};
			}
		}
		$row->{'customize'} = 1 unless $row->{'ignore'};
		row_defaults($row, $key);
		row_params($row);
		push(@fields, $row);
	}
	return \@fields;
}

sub row_defaults {
	my ($row, $key) = @_;
	$row->{'label'} = $key unless defined $row->{'label'};
	$row->{'type'} ||= 'text';
	$row->{'width'} ||= 591;
	$row->{'rows'} ||= 10;
	$row->{'date_order'} ||= 'mdy';
	$row->{'date_time'} ||= 'hhmm';
	$row->{'date_minutes'} ||= '5';
	$row->{'date_y_start'} ||= '2000';
	$row->{'date_y_end'} ||= '2008';
	$row->{'filenames'} ||= 'dirify';
	$row->{'show_buttons'} = 1 unless defined $row->{'show_buttons'};
	$row->{'text_filters'} = 1 unless defined $row->{'text_filters'};
}

sub row_params {
	my ($row) = @_;
	for my $key (qw( type date_order date_time date_minutes filenames )) {
		next unless $row->{$key};
		$row->{$row->{$key} . '_selected'} = 1;
	}
	$row->{'show_minutes'} = ($row->{'date_time'} ne 'hh') ? 1 : 0;
	if ($row->{'type'} =~ /^text/) {
		$row->{'show_text'} = 1;
	}
	if ($row->{'type'} =~ /^menu|radio$/) {
		$row->{'show_choices'} = 1;
	}
}

sub using_default {
# determine whether default settings are in use for a given blog
# and data type
	my ($q, $type, $data) = @_;
	my $has_data;
	if (ref($data) eq 'HASH') {
		$has_data = %$data ? 1 : 0;
	} elsif (ref($data) eq 'ARRAY') {
		$has_data = @$data ? 1 : 0;
	} else {
		$has_data = $data ? 1 : 0;
	}
	if (($q->param('blog_id') != -1) && !$has_data) {
		my $default_key = config_key(-1, $type);
			# we have default settings and no blog-specific settings
		if (load_plugindata($default_key)) {
			return 1;
		}
	}
	return 0;
}

sub blogs_for_author {
	my ($app, $selected) = @_;
	$selected ||= 0;
	my $author = $app->{'author'};
	return [] unless $author;
	require MT::Permission;
	my @blogs = ();
	for my $perms (MT::Permission->load({ 'author_id' => $author->id })) {
		next unless $author->is_superuser() || $perms->can_administer_blog;
		my $blog = MT::Blog->load($perms->blog_id);
		my %selected = ($blog->id == $selected) ? ('selected' => 1) : ();
		push(@blogs, { 'id' => $blog->id, 'name' => $blog->name, %selected });
	}
	@blogs = sort { $a->{'name'} cmp $b->{'name'} } @blogs;
	return \@blogs;
}

sub uri {
	my ($app) = shift;
	return $app->app_uri(@_);
}

package DatasourceHolder;
# dummy object class that we can pass to driver's column_defs() to get
# column defs for an arbitrary datasource
my $datasource;
sub new {
	bless({}, shift);
}
sub properties {
	return { 'datasource' => $datasource };
}
sub datasource {
	$datasource = $_[0];
}
sub column_defs {
# need to define this for sqlite; it's not used for anything we're
# concerned with, so just return a dummy
	return {};
}
1;
